﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text.RegularExpressions;

namespace HeuristicLab.BioBoost.Utils {

  [StorableClass]
  [Item("StepFunction", "A simple step function specifying upper bounds and corresponding values.")]
  public class StepFunction : ParameterizedNamedItem {

    #region parameters
    public IValueLookupParameter<DoubleMatrix> ValuesParameter {
      get { return (IValueLookupParameter<DoubleMatrix>) Parameters["Values"]; }
    }
    public DoubleMatrix Values {
      get { return ValuesParameter.Value; }
      set { ValuesParameter.Value = value; }
    }
    #endregion

    #region Construction & Cloning
    public StepFunction() {
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("Values", "The upper bounds and corresponding values.", new DoubleMatrix(0, 2, new[] {"Upper Bound", "Value"})));
    }
    [StorableConstructor]
    protected StepFunction(bool isDeserializing) : base(isDeserializing) {}
    protected StepFunction(StepFunction orig, Cloner cloner) : base(orig, cloner) {}
    public override IDeepCloneable Clone(Cloner cloner) {
      return new StepFunction(this, cloner);
    }
    #endregion

    public void Sort() {
      if (Values.Columns != 2)
        throw new InvalidOperationException("StepFunction values need two columns (upper bound and value)");
      var steps = Enumerable.Range(0, Values.Rows).Select(i => new {s = Values[i, 0], v = Values[i, 1]}).OrderBy(p => p.s).ToList();
      for (int i = 0; i<Values.Rows; i++) {
        Values[i, 0] = steps[i].s;
        Values[i, 1] = steps[i].v;
      }
    }

    public double GetValue(double x) {
      int i = 0;
      while (i < Values.Rows-1 && x > Values[i, 0]) i++;
      return Values[i, 1];
    }

    public static StepFunction Parse(String s, String rowSepRegex="/", String colSepRegex="->") {
      var values = new List<Tuple<double, double>>();
      foreach (var row in Regex.Split(s, rowSepRegex)) {
        var cols = Regex.Split(row, colSepRegex);
        if (cols.Length != 2) {
          throw new ArgumentException("step function must contain two rows (separated by /" + colSepRegex + "/");
        }
        try {
          values.Add(new Tuple<double, double>(ParseDouble(cols[0]), ParseDouble(cols[1])));
        }
        catch (Exception x) {
          if (x is FormatException || x is ArgumentNullException || x is OverflowException) {
            throw new ArgumentException(String.Format("could not parse row \"{0}\"", row), x);
          }
        }
      }
      var sf = new StepFunction();
      var matrix = new DoubleMatrix(values.Count, 2, new[] {"Upper Bound", "Value"});
      for (int i = 0; i < values.Count; i++) {
        matrix[i, 0] = values[i].Item1;
        matrix[i, 1] = values[i].Item2;
      }
      sf.Values = matrix;
      sf.Sort();
      return sf;
    }


    public static double ParseDouble(String s) {
      return Double.Parse(s, NumberStyles.Float, NumberFormatInfo.InvariantInfo);
    }
  }
}
